import logging
import sys

from loguru import logger


class LoguruHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
            if record.levelno <= logging.INFO:
                level = {"DEBUG": "TRACE", "INFO": "DEBUG"}.get(level, level)
        except ValueError:
            level = record.levelno

        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def get_subclasses(cls):
    yield from cls.__subclasses__()
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)



from __future__ import annotations

import re
from collections.abc import Sequence
from typing import ClassVar

from graia.amnesia.message import MessageChain
from graia.amnesia.message.element import Element


class Formatter:
    format_string: str
    __message_chain_class__: ClassVar[type[MessageChain]] = MessageChain

    def __init__(self, format_string: str) -> None:
        self.format_string = format_string

    @classmethod
    def ensure_element(cls, obj: str | Element) -> Element:
        return cls.__message_chain_class__._text_class(obj) if isinstance(obj, str) else obj

    @classmethod
    def extract_chain(cls, obj: Element | MessageChain | str | Sequence[Element]) -> list[Element]:
        if isinstance(obj, MessageChain):
            return obj.content
        if isinstance(obj, str):
            obj = cls.ensure_element(obj)
        if isinstance(obj, Element):
            return [obj]
        if isinstance(obj, Sequence):
            return list(obj)

    def format(self, *o_args: Element | MessageChain | str, **o_kwargs: Element | MessageChain | str) -> MessageChain:
        args: list[list[Element]] = [self.extract_chain(e) for e in o_args]
        kwargs: dict[str, list[Element]] = {k: self.extract_chain(e) for k, e in o_kwargs.items()}

        args_mapping: dict[str, list[Element]] = {f"\x02{index}\x02": chain for index, chain in enumerate(args)}
        kwargs_mapping: dict[str, list[Element]] = {f"\x03{key}\x03": chain for key, chain in kwargs.items()}

        result = self.format_string.format(*args_mapping, **{k: f"\x03{k}\x03" for k in kwargs})

        chain_list: list[Element] = []

        for i in re.split("([\x02\x03][\\d\\w]+[\x02\x03])", result):
            if match := re.fullmatch("(?P<header>[\x02\x03])(?P<content>\\w+)(?P=header)", i):
                header = match["header"]
                full: str = match[0]
                if header == "\x02":  # from args
                    chain_list.extend(args_mapping[full])
                else:  # \x03, from kwargs
                    chain_list.extend(kwargs_mapping[full])
            else:
                chain_list.append(self.ensure_element(i))
        return MessageChain(chain_list).merge()



from __future__ import annotations

import asyncio
from datetime import timedelta
from heapq import heappop, heappush
from time import time
from typing import Any

from launart import Launart, Service
from launart.status import Phase


class Memcache:
    cache: dict[str, tuple[float | None, Any]]
    expire: list[tuple[float, str]]

    def __init__(
        self,
        cache: dict[str, tuple[float | None, Any]],
        expire: list[tuple[float, str]],
    ):
        self.cache = cache
        self.expire = expire

    async def get(self, key: str, default: Any = None) -> Any:
        value = self.cache.get(key)
        if value:
            if value[0] is None or value[0] >= time():
                return value[1]
            del self.cache[key]
        return default

    async def set(self, key: str, value: Any, expire: timedelta | None = None) -> None:
        if expire is None:
            self.cache[key] = (None, value)
        else:
            expire_time = time() + expire.total_seconds()
            self.cache[key] = (expire_time, value)
            heappush(self.expire, (expire_time, key))

    async def delete(self, key: str, strict: bool = False) -> None:
        if strict or key in self.cache:
            del self.cache[key]

    async def clear(self) -> None:
        self.cache.clear()
        self.expire.clear()

    async def has(self, key: str) -> bool:
        return key in self.cache

    async def keys(self) -> list[str]:
        return list(self.cache.keys())


class MemcacheService(Service):
    id = "cache.client/memcache"

    interval: float
    _cache: dict[str, tuple[float | None, Any]]
    expire: list[tuple[float, str]]

    def __init__(
        self,
        interval: float = 0.1,
        cache: dict[str, tuple[float | None, Any]] | None = None,
        expire: list[tuple[float, str]] | None = None,
    ):
        self.interval = interval
        self._cache = cache or {}
        self.expire = expire or []
        super().__init__()

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"blocking"}

    @property
    def cache(self):
        return Memcache(self._cache, self.expire)

    async def launch(self, manager: Launart) -> None:
        async with self.stage("blocking"):
            while not manager.status.exiting:
                while self.expire and self.expire[0][0] <= time():
                    _, key = heappop(self.expire)
                    self._cache.pop(key, None)
                await asyncio.sleep(self.interval)



from __future__ import annotations

import asyncio
import logging

from launart import Launart, Service
from launart.status import Phase
from launart.utilles import any_completed
from loguru import logger

try:
    from uvicorn import Config, Server
except ImportError:
    raise ImportError(
        "dependency 'uvicorn' is required for asgi service\nplease install it or install 'graia-amnesia[asgi]'"
    )

from . import asgitypes
from .middleware import DispatcherMiddleware


async def _empty_asgi_handler(scope, receive, send):
    if scope["type"] == "lifespan":
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                await send({"type": "lifespan.startup.complete"})
                return
            elif message["type"] == "lifespan.shutdown":
                await send({"type": "lifespan.shutdown.complete"})
                return

    await send(
        {
            "type": "http.response.start",
            "status": 404,
            "headers": [(b"content-length", b"0")],
        }
    )
    await send({"type": "http.response.body"})


class LoguruHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)

        frame, depth = logging.currentframe(), 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )


class WithoutSigHandlerServer(Server):
    def install_signal_handlers(self) -> None:
        pass


class UvicornASGIService(Service):
    id = "asgi.service/uvicorn"

    middleware: DispatcherMiddleware
    host: str
    port: int

    def __init__(
        self,
        host: str,
        port: int,
        mounts: dict[str, asgitypes.ASGI3Application] | None = None,
    ):
        self.host = host
        self.port = port
        self.middleware = DispatcherMiddleware(mounts or {"\0\0\0": _empty_asgi_handler})
        super().__init__()

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def launch(self, manager: Launart) -> None:
        async with self.stage("preparing"):
            self.server = WithoutSigHandlerServer(
                Config(self.middleware, host=self.host, port=self.port, factory=False)
            )

            level = logging.getLevelName(20)  # default level for uvicorn
            logging.basicConfig(handlers=[LoguruHandler()], level=level)
            PATCHES = ["uvicorn.error", "uvicorn.asgi", "uvicorn.access", ""]
            for name in PATCHES:
                target = logging.getLogger(name)
                target.handlers = [LoguruHandler(level=level)]
                target.propagate = False

            serve_task = asyncio.create_task(self.server.serve())

        async with self.stage("blocking"):
            await any_completed(serve_task, manager.status.wait_for_sigexit())

        async with self.stage("cleanup"):
            logger.warning("try to shutdown uvicorn server...")
            self.server.should_exit = True
            await any_completed(serve_task, asyncio.sleep(5))
            if not serve_task.done():
                logger.warning("timeout, force exit uvicorn server...")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from . import MessageChain


class Element:
    _chain_class: type[MessageChain]

    def __str__(self) -> str:
        return ""

    def __add__(self: Element, content: MessageChain | list[Element] | Element | str) -> MessageChain:
        from . import MessageChain

        if isinstance(content, str):
            content = [self._chain_class._text_class(content)]
        if isinstance(content, Element):
            content = [content]
        if isinstance(content, MessageChain):
            content = content.content
        return self._chain_class(content + [self])

    def __radd__(self: Element, content: MessageChain | list[Element] | Element | str) -> MessageChain:
        from . import MessageChain

        if isinstance(content, str):
            content = [self._chain_class._text_class(content)]
        if isinstance(content, Element):
            content = [content]
        if isinstance(content, MessageChain):
            content = content.content
        return self._chain_class([self] + content)

    def __eq__(self, other):
        if not isinstance(other, Element):
            return False
        return vars(self) == vars(other)


class Text(Element):
    text: str
    style: str | None

    def __init__(self, text: str, style: str | None = None) -> None:
        """实例化一个 Text 消息元素, 用于承载消息中的文字.

        Args:
            text (str): 元素所包含的文字
            style (Optional[str]): 默认为空, 文字的样式
        """
        self.text = text
        self.style = style

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"Text(text={self.text}{f', style={self.style}' if self.style else ''})"


class Unknown(Element):
    type: str
    raw_data: Any

    def __init__(self, type: str, raw_data: Any) -> None:
        self.type = type
        self.raw_data = raw_data

    def __str__(self) -> str:
        return f"[$Unknown:type={self.type}]"

    def __repr__(self) -> str:
        return f"Unknown(type={self.type}, raw=<{self.raw_data.__class__.__name__}>)"



from .chain import MessageChain as MessageChain
from .element import Element as Element
from .element import Text as Text
from .element import Unknown as Unknown
from .formatter import Formatter as Formatter

Element._chain_class = MessageChain
MessageChain._text_class = Text



from typing import TYPE_CHECKING, Any, Callable, ClassVar

from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.schema import Table

_NAMING_CONVENTION = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


_callbacks = []


def register_callback(callback: Callable[[type["Base"], dict[str, Any]], Any]) -> None:
    """
    Register a callback to be called when a new Base subclass is created.
    """
    _callbacks.append(callback)


def remove_callback(callback: Callable[[type["Base"], dict[str, Any]], Any]) -> None:
    """
    Remove a previously registered callback.
    """
    if callback in _callbacks:
        _callbacks.remove(callback)


class Base(AsyncAttrs, DeclarativeBase):
    __abstract__ = True
    metadata = MetaData(naming_convention=_NAMING_CONVENTION)

    if TYPE_CHECKING:
        __table__: ClassVar[Table]  # type: ignore

    def __init_subclass__(cls, **kwargs):
        for callback in _callbacks:
            callback(cls, kwargs)

        if not hasattr(cls, "__tablename__") and "tablename" in kwargs:
            cls.__tablename__ = kwargs["tablename"]
        if not hasattr(cls, "__table_args__") and "table_args" in kwargs:
            cls.__table_args__ = kwargs["table_args"]
        if not hasattr(cls, "__mapper__") and "mapper" in kwargs:
            cls.__mapper__ = kwargs["mapper"]
        if not hasattr(cls, "__mapper_args__") and "mapper_args" in kwargs:
            cls.__mapping_args__ = kwargs["mapper_args"]

        super().__init_subclass__(**kwargs)

        if not hasattr(cls, "__table__"):
            return

        bind_key: str | None = getattr(cls, "__bind_key__", kwargs.get("bind_key", None))

        if bind_key is None:
            bind_key = ""

        cls.__table__.info["bind_key"] = bind_key



from __future__ import annotations

from typing import cast

from launart import Launart, Service
from launart.status import Phase

try:
    from aiohttp import ClientSession, ClientTimeout
except ImportError:
    raise ImportError(
        "dependency 'aiohttp' is required for aiohttp client service\n"
        "please install it or install 'graia-amnesia[aiohttp]'"
    )


class AiohttpClientService(Service):
    id = "http.client/aiohttp"
    session: ClientSession

    def __init__(self, session: ClientSession | None = None) -> None:
        self.session = cast(ClientSession, session)
        super().__init__()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "cleanup"}

    @property
    def required(self):
        return set()

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            if self.session is None:
                self.session = ClientSession(timeout=ClientTimeout(total=None))
        async with self.stage("cleanup"):
            await self.session.close()
